
public class Tableaux {

	static void afficher(int[] tab) {
		for (int i = 0; i < tab.length; i++)
			System.out.print(tab[i] + " ");
		System.out.println();
	}

	static String conversionChaine(int[] tab) {
		String resultat = "[";
		boolean first = true;
		for (int i = 0; i < tab.length; i++) {
			if (!first)
				resultat += ";";
			resultat += tab[i];
			first = false;
		}
		resultat += "]";
		return resultat;
	}

	// Ici le critère indique si :
	// -1 : on cherche le min
	// 0 : on cherche la valeur donnée
	// 1 : on cherche le max
	static int indiceElement(int[] tab, int critere, int valeur) {
		if (tab.length < 1)
			return -1;
		int indice = -1;
		boolean resultat = false;
		for (int i = 0; i < tab.length; i++) {
			if (indice != -1) {
				switch (critere) {
				case -1:
					resultat = tab[indice] > tab[i];
					break;
				case 0:
					resultat = false;
					break;
				case 1:
					resultat = tab[indice] < tab[i];
					break;
				default:
					System.err.println("Bug, valeur de critère " + critere + " invalide dans indiceElement");
				}
			} else {
				resultat = (critere != 0) || (tab[i] == valeur);
			}
			if (resultat)
				indice = i;
		}
		return indice;
	}

	static int min(int[] tab) {
		return tab[indiceMin(tab)];
	}

	static int max(int[] tab) {
		return tab[indiceMax(tab)];
	}

	static int indiceMin(int[] tab) {
		return indiceElement(tab, -1, 0);
	}

	static int indiceMax(int[] tab) {
		return indiceElement(tab, 1, 0);
	}

	static int indiceDe(int[] tab, int val) {
		return indiceElement(tab, 0, val);
	}

	static int[] listeIndicesMin(int[] tab) {
		int nombreIndices;
		int[] resultat;
		int min;

		if (tab.length == 0)
			return null;
		min = tab[0];
		nombreIndices = 1;
		for (int i = 1; i < tab.length; i++)
			if (tab[i] <= min) {
				if (tab[i] == min) {
					nombreIndices++;
				} else {
					min = tab[i];
					nombreIndices = 1;
				}
			}

		resultat = new int[nombreIndices];
		int j = 0;
		for (int i = 0; i < tab.length; i++)
			if (tab[i] == min) {
				resultat[j] = i;
				j++;
			}
		return resultat;
	}

	static boolean estCroissant(int[] tab) {
		for (int i = 1; i < tab.length; i++)
			if (tab[i - 1] > tab[i])
				return false;
		return true;
	}

	static int occurrenceMax(int[] tab) {
		// Ici, pour l'exemple, on suppose que tous les éléments ont une valeur entre 0
		// et 100
		int[] occurences;
		int max = 0, maxOccurences = 0;

		occurences = new int[101];
		for (int i = 0; i < tab.length; i++) {
			if ((tab[i] < 0) || (tab[i] > 100))
				return 0;
			occurences[tab[i]]++;
			if ((occurences[tab[i]] > maxOccurences) || ((occurences[tab[i]] == maxOccurences) && (tab[i] < max))) {
				max = tab[i];
				maxOccurences = occurences[tab[i]];
			}
		}
		return max;
	}

	static boolean palindromeRec(int[] tab, int inf, int sup) {
		if (inf >= sup) {
			return true;
		} else {
			if (tab[inf] != tab[sup])
				return false;
			else
				return palindromeRec(tab, inf + 1, sup - 1);
		}
	}

	static boolean palindrome(int[] tab) {
		return palindromeRec(tab, 0, tab.length - 1);
	}

	static int retentionEau(int[] tab) {
		final int GAUCHE = 0;
		final int DROITE = 1;
		int[] indice, hauteur;
		int cote, prochain;
		int eau = 0;

		if (tab.length == 0)
			return 0;
		indice = new int[2];
		hauteur = new int[2];
		indice[GAUCHE] = 0;
		indice[DROITE] = tab.length - 1;
		hauteur[GAUCHE] = tab[0];
		hauteur[DROITE] = tab[tab.length - 1];

		while (indice[DROITE] - indice[GAUCHE] > 1) {
			if (hauteur[GAUCHE] < hauteur[DROITE]) {
				cote = GAUCHE;
				prochain = indice[GAUCHE] + 1;
			} else {
				cote = DROITE;
				prochain = indice[DROITE] - 1;
			}
			if (tab[prochain] < hauteur[cote]) {
				eau += hauteur[cote] - tab[prochain];
			} else {
				hauteur[cote] = tab[prochain];
			}
			indice[cote] = prochain;
		}
		return eau;
	}

	// Solution en diviser pour régner
	// On ne demandait pas une solution aussi élaborée, une recherche pour tous les intervalles est acceptable
	static int sousTableauDeSommeMax(int[] T) {
		return sousTableauDeSommeMaxRec(T, 0, T.length-1);
	}

	static int sousTableauDeSommeMaxRec(int[] T, int deb, int fin) {
		if (deb > fin)
			return 0;
		if (deb == fin)
			return Math.max(0, T[deb]);
		int milieu = (deb+fin) / 2;
		int somme = 0;
		int max = 0, maxGauche = 0, maxDroite = 0;
		// Soit le sous tableau passe par le milieu
		for (int i=milieu+1; i<=fin; i++) {
			somme += T[i];
			maxDroite = Math.max(maxDroite, somme);
		}
		somme = 0;
		for (int i=milieu; i>=deb; i--) {
			somme += T[i];
			maxGauche = Math.max(maxGauche, somme);
		}
		max = maxGauche + maxDroite;
		// Soit il est complètement à gauche
		max = Math.max(max, sousTableauDeSommeMaxRec(T, deb, milieu));
		// Soit il est complètement à droite
		max = Math.max(max, sousTableauDeSommeMaxRec(T, milieu+1, fin));
		return max;
	}
	
	// Solution par programmation dynamique
	// On ne demandait pas une solution aussi élaborée, une énumération de toutes les sous séquence est acceptable
	static int sousSequenceCroissante(int [] T) {
		// T[i] = longueur de la plus longue sous séquence croissante pour les éléments de 0 à i
		int [] longueur = new int[T.length];
		int max = Math.min(1, T.length);

		for (int i=0; i<T.length; i++) {
			// Sequence ne contenant que l'élément
			longueur[i] = 1;
			for (int j=0; j<i; j++) {
				if (T[j] < T[i]) {
					// Séquence la plus longue trouvé jusqu'alors ou séquence contenant l'élément T[j] puis T[i]
					longueur[i] = Math.max(longueur[i], longueur[j]+1);
					max = Math.max(max, longueur[i]);
				}
			}
		}
		return max;
	}
}